package ru.dbotthepony.mc.otm.data

import com.google.gson.GsonBuilder
import com.google.gson.JsonArray
import com.google.gson.JsonElement
import com.google.gson.JsonObject
import com.mojang.datafixers.util.Either
import com.mojang.serialization.Codec
import com.mojang.serialization.JsonOps
import com.mojang.serialization.codecs.RecordCodecBuilder
import net.minecraft.core.registries.BuiltInRegistries
import net.minecraft.core.registries.Registries
import net.minecraft.resources.ResourceLocation
import net.minecraft.server.packs.resources.ResourceManager
import net.minecraft.server.packs.resources.SimpleJsonResourceReloadListener
import net.minecraft.tags.TagKey
import net.minecraft.util.profiling.ProfilerFiller
import net.minecraft.world.level.block.Block
import net.minecraft.world.level.block.state.BlockState
import net.neoforged.neoforge.event.AddReloadListenerEvent
import org.apache.logging.log4j.LogManager
import ru.dbotthepony.mc.otm.core.math.Decimal
import ru.dbotthepony.mc.otm.core.registryName
import ru.dbotthepony.mc.otm.data.codec.DecimalCodec
import ru.dbotthepony.mc.otm.data.codec.minRange

object FlywheelMaterials : SimpleJsonResourceReloadListener(GsonBuilder().setPrettyPrinting().create(), "otm_flywheel_materials") {
	data class Entry(
		val id: Either<TagKey<Block>, Block>,
		val storage: Decimal,
		val priority: Int = 0,
		val receiveEfficiency: Decimal = Decimal.ONE,
		val momentumLossSpeed: Decimal = Decimal.ONE
	) : Comparable<Entry> {
		constructor(id: Block, storage: Decimal, priority: Int = 0, windUpEfficiency: Decimal = Decimal.ONE, windDownEfficiency: Decimal = Decimal.ONE) : this(Either.right(id), storage, priority, windUpEfficiency, windDownEfficiency)
		constructor(id: TagKey<Block>, storage: Decimal, priority: Int = 0, windUpEfficiency: Decimal = Decimal.ONE, windDownEfficiency: Decimal = Decimal.ONE) : this(Either.left(id), storage, priority, windUpEfficiency, windDownEfficiency)

		override fun compareTo(other: Entry): Int {
			val selfLeft = id.left().isPresent
			val otherLeft = other.id.left().isPresent

			if (selfLeft && otherLeft || !selfLeft && !otherLeft) {
				val cmp = priority.compareTo(other.priority)

				if (cmp == 0) {
					if (selfLeft) {
						return id.left().get().location.compareTo(other.id.left().get().location)
					} else {
						return id.right().get().registryName!!.compareTo(other.id.right().get().registryName!!)
					}
				}

				return cmp
			} else if (selfLeft) {
				return -1
			} else {
				return 1
			}
		}
	}

	private var elements = ArrayList<Entry>()
	private var mapped = HashMap<Block, Entry>()
	private var mappingChanged = false

	private fun mapping(): Map<Block, Entry> {
		var mapped = mapped

		if (mappingChanged) {
			mapped = HashMap()

			for (element in elements) {
				element.id
					.ifLeft {
						for (b in BuiltInRegistries.BLOCK.getTagOrEmpty(it)) {
							mapped[b.value()] = element
						}
					}
					.ifRight {
						mapped[it] = element
					}
			}

			this.mapped = mapped
			this.mappingChanged = false
			return mapped
		}

		return mapped
	}

	operator fun get(index: Block): Entry? {
		return mapping()[index]
	}

	operator fun get(index: BlockState): Entry? {
		return this[index.block]
	}

	operator fun contains(value: Block): Boolean {
		return value in mapping()
	}

	operator fun contains(value: BlockState): Boolean {
		return value.block in mapping()
	}

	override fun apply(
		map: MutableMap<ResourceLocation, JsonElement>,
		manager: ResourceManager,
		profiler: ProfilerFiller
	) {
		val elements = ArrayList<Entry>()

		for ((path, element) in map.entries) {
			if (element is JsonObject) {
				CODEC.decode(JsonOps.INSTANCE, element)
					.ifError { LOGGER.error("Unable to deserialize flywheel material entry at $path: $it") }
					.ifSuccess { elements.add(it.first) }
			} else if (element is JsonArray) {
				for ((i, elem) in element.withIndex()) {
					CODEC.decode(JsonOps.INSTANCE, elem)
						.ifError { LOGGER.error("Unable to deserialize flywheel material entry at $path, offset $i: $it") }
						.ifSuccess { elements.add(it.first) }
				}
			}
		}

		elements.sortDescending()

		this.elements = elements
		this.mappingChanged = true
	}

	val CODEC: Codec<Entry> = RecordCodecBuilder.create {
		it.group(
			Codec.either(
				TagKey.hashedCodec(Registries.BLOCK),
				BuiltInRegistries.BLOCK.byNameCodec()
			).fieldOf("id").forGetter(Entry::id),
			DecimalCodec.minRange(Decimal.ONE).fieldOf("storage").forGetter(Entry::storage),
			Codec.INT.optionalFieldOf("priority", 0).forGetter(Entry::priority),
			DecimalCodec.minRange(Decimal.ZERO).optionalFieldOf("receive_efficiency", Decimal.ONE).forGetter(Entry::receiveEfficiency),
			DecimalCodec.minRange(Decimal.ZERO).optionalFieldOf("momentum_loss_speed", Decimal.ONE).forGetter(Entry::momentumLossSpeed),
		).apply(it, ::Entry)
	}

	fun reloadEvent(event: AddReloadListenerEvent) {
		event.addListener(this)
	}

	private val LOGGER = LogManager.getLogger()
}
